/**
 * Extracted from the following sources:
 *
 * | License | Name           | Copyright                                  |
 * |---------|----------------|--------------------------------------------|
 * | MIT     | sha.js         | (c) 2013-2018 sha.js contributors          |
 * | MIT     | Crypto-js      | (c) 2009-2013 Jeff Mott.                  |
 * | MIT     | browserify-aes | (c) 2014-2017 browserify-aes contributors  |
 */
import { Buffer } from "buffer"

function bufferXor(a: Buffer, b: Buffer) {
  const length = Math.min(a.length, b.length)
  const buffer = Buffer.alloc(length)

  for (let i = 0; i < length; ++i) {
    buffer[i] = a[i] ^ b[i]
  }

  return buffer
}

function toUInt32Array(buf: Buffer) {
  const len = (buf.length / 4) | 0
  const out: number[] = new Array(len)

  for (let i = 0; i < len; i++) {
    out[i] = buf.readUInt32BE(i * 4)
  }

  return out
}

function cryptBlock(
  M: number[],
  keySchedule: number[],
  subMix: number[][],
  sbox: number[],
  nRounds: number
) {
  const [subMix_0, subMix_1, subMix_2, subMix_3] = subMix

  let s0 = M[0] ^ keySchedule[0]
  let s1 = M[1] ^ keySchedule[1]
  let s2 = M[2] ^ keySchedule[2]
  let s3 = M[3] ^ keySchedule[3]
  let ksRow = 4

  for (let round = 1; round < nRounds; round++) {
    const t0 =
      subMix_0[s0 >>> 24] ^
      subMix_1[(s1 >>> 16) & 0xff] ^
      subMix_2[(s2 >>> 8) & 0xff] ^
      subMix_3[s3 & 0xff] ^
      keySchedule[ksRow++]
    const t1 =
      subMix_0[s1 >>> 24] ^
      subMix_1[(s2 >>> 16) & 0xff] ^
      subMix_2[(s3 >>> 8) & 0xff] ^
      subMix_3[s0 & 0xff] ^
      keySchedule[ksRow++]
    const t2 =
      subMix_0[s2 >>> 24] ^
      subMix_1[(s3 >>> 16) & 0xff] ^
      subMix_2[(s0 >>> 8) & 0xff] ^
      subMix_3[s1 & 0xff] ^
      keySchedule[ksRow++]
    const t3 =
      subMix_0[s3 >>> 24] ^
      subMix_1[(s0 >>> 16) & 0xff] ^
      subMix_2[(s1 >>> 8) & 0xff] ^
      subMix_3[s2 & 0xff] ^
      keySchedule[ksRow++]
    s0 = t0
    s1 = t1
    s2 = t2
    s3 = t3
  }

  let t0 =
    ((sbox[s0 >>> 24] << 24) |
      (sbox[(s1 >>> 16) & 0xff] << 16) |
      (sbox[(s2 >>> 8) & 0xff] << 8) |
      sbox[s3 & 0xff]) ^
    keySchedule[ksRow++]
  let t1 =
    ((sbox[s1 >>> 24] << 24) |
      (sbox[(s2 >>> 16) & 0xff] << 16) |
      (sbox[(s3 >>> 8) & 0xff] << 8) |
      sbox[s0 & 0xff]) ^
    keySchedule[ksRow++]
  let t2 =
    ((sbox[s2 >>> 24] << 24) |
      (sbox[(s3 >>> 16) & 0xff] << 16) |
      (sbox[(s0 >>> 8) & 0xff] << 8) |
      sbox[s1 & 0xff]) ^
    keySchedule[ksRow++]
  let t3 =
    ((sbox[s3 >>> 24] << 24) |
      (sbox[(s0 >>> 16) & 0xff] << 16) |
      (sbox[(s1 >>> 8) & 0xff] << 8) |
      sbox[s2 & 0xff]) ^
    keySchedule[ksRow++]
  t0 = t0 >>> 0
  t1 = t1 >>> 0
  t2 = t2 >>> 0
  t3 = t3 >>> 0

  return [t0, t1, t2, t3]
}

// AES constants
const RCON = [0x00, 0x01, 0x02, 0x04, 0x08, 0x10, 0x20, 0x40, 0x80, 0x1b, 0x36]
const G = (() => {
  // Compute double table
  const d = new Array<number>(256)
  for (let j = 0; j < 256; j++) {
    if (j < 128) {
      d[j] = j << 1
    } else {
      d[j] = (j << 1) ^ 0x11b
    }
  }

  const sbox: number[] = []
  const invSBox: number[] = []
  const subMix: number[][] = [[], [], [], []]
  const invSubMix: number[][] = [[], [], [], []]

  // Walk GF(2^8)
  let x = 0
  let xi = 0
  for (let i = 0; i < 256; ++i) {
    // Compute sbox
    let sx = xi ^ (xi << 1) ^ (xi << 2) ^ (xi << 3) ^ (xi << 4)
    sx = (sx >>> 8) ^ (sx & 0xff) ^ 0x63
    sbox[x] = sx
    invSBox[sx] = x

    // Compute multiplication
    const x2 = d[x]
    const x4 = d[x2]
    const x8 = d[x4]

    // Compute sub bytes, mix columns tables
    let t = (d[sx] * 0x101) ^ (sx * 0x1010100)
    subMix[0][x] = (t << 24) | (t >>> 8)
    subMix[1][x] = (t << 16) | (t >>> 16)
    subMix[2][x] = (t << 8) | (t >>> 24)
    subMix[3][x] = t

    // Compute inv sub bytes, inv mix columns tables
    t = (x8 * 0x1010101) ^ (x4 * 0x10001) ^ (x2 * 0x101) ^ (x * 0x1010100)
    invSubMix[0][sx] = (t << 24) | (t >>> 8)
    invSubMix[1][sx] = (t << 16) | (t >>> 16)
    invSubMix[2][sx] = (t << 8) | (t >>> 24)
    invSubMix[3][sx] = t

    if (x === 0) {
      x = xi = 1
    } else {
      x = x2 ^ d[d[d[x8 ^ x2]]]
      xi ^= d[d[xi]]
    }
  }

  return {
    sbox,
    invSBox,
    subMix,
    invSubMix,
  }
})()

class AES {
  private key: number[]
  private nRounds!: number
  private invKeySchedule!: number[]

  constructor(key: Buffer) {
    this.key = toUInt32Array(key)
    this.reset()
  }

  private reset() {
    const keyWords = this.key
    const keySize = keyWords.length
    const nRounds = keySize + 6
    const ksRows = (nRounds + 1) * 4

    const keySchedule: number[] = []
    for (let k = 0; k < keySize; k++) {
      keySchedule[k] = keyWords[k]
    }

    for (let k = keySize; k < ksRows; k++) {
      let t = keySchedule[k - 1]

      if (k % keySize === 0) {
        t = (t << 8) | (t >>> 24)
        t =
          (G.sbox[t >>> 24] << 24) |
          (G.sbox[(t >>> 16) & 0xff] << 16) |
          (G.sbox[(t >>> 8) & 0xff] << 8) |
          G.sbox[t & 0xff]

        t ^= RCON[(k / keySize) | 0] << 24
      } else if (keySize > 6 && k % keySize === 4) {
        t =
          (G.sbox[t >>> 24] << 24) |
          (G.sbox[(t >>> 16) & 0xff] << 16) |
          (G.sbox[(t >>> 8) & 0xff] << 8) |
          G.sbox[t & 0xff]
      }

      keySchedule[k] = keySchedule[k - keySize] ^ t
    }

    const invKeySchedule = []
    for (let ik = 0; ik < ksRows; ik++) {
      const ksR = ksRows - ik
      const tt = keySchedule[ksR - (ik % 4 ? 0 : 4)]

      if (ik < 4 || ksR <= 4) {
        invKeySchedule[ik] = tt
      } else {
        invKeySchedule[ik] =
          G.invSubMix[0][G.sbox[tt >>> 24]] ^
          G.invSubMix[1][G.sbox[(tt >>> 16) & 0xff]] ^
          G.invSubMix[2][G.sbox[(tt >>> 8) & 0xff]] ^
          G.invSubMix[3][G.sbox[tt & 0xff]]
      }
    }

    this.nRounds = nRounds
    this.invKeySchedule = invKeySchedule
  }

  decryptBlock(buffer: Buffer) {
    const M = toUInt32Array(buffer)

    // swap
    const m1 = M[1]
    M[1] = M[3]
    M[3] = m1

    const out = cryptBlock(M, this.invKeySchedule, G.invSubMix, G.invSBox, this.nRounds)

    const buf = Buffer.allocUnsafe(16)
    buf.writeUInt32BE(out[0], 0)
    buf.writeUInt32BE(out[3], 4)
    buf.writeUInt32BE(out[2], 8)
    buf.writeUInt32BE(out[1], 12)
    return buf
  }

  static blockSize = 4 * 4
  static keySize = 256 / 8
}

function splitter() {
  let cache = Buffer.allocUnsafe(0)
  return {
    add(data: Buffer) {
      cache = Buffer.concat([cache, data])
      return this
    },
    get() {
      if (cache.length >= 16) {
        const out = cache.slice(0, 16)
        cache = cache.slice(16)
        return out
      }
      return null
    },
  }
}

// AES-256-CBC
// == createDecipheriv("aes-256-cbc", key, iv).setAutoPadding(false).update(data)
export function decryptData(key: Buffer, iv: Buffer, data: Buffer) {
  if (iv.length !== 16) {
    throw new TypeError(`invalid iv length ${iv.length}`)
  }
  if (key.length !== 32) {
    throw new TypeError(`invalid key length ${key.length}`)
  }

  const cipher = new AES(key)
  let prev = Buffer.from(iv)
  const cache = splitter().add(data)
  let chunk: Buffer | null
  const res: Buffer[] = []
  while ((chunk = cache.get())) {
    const pad = prev
    prev = chunk
    const out = cipher.decryptBlock(chunk)
    res.push(bufferXor(out, pad))
  }
  return Buffer.concat(res)
}
