import { Buffer } from "buffer"
import type { Crypto } from "./crypto"
import { invariant } from "../errors"

type integer = number

export interface AttachmentMetadata {
  itemUUID: string
  contentSize: integer
  external: boolean
  updatedAt: integer
  txTimestamp: integer
  /** Encoded as base64 encoded OPData */
  overview: {
    filename: string
  }
  createdAt: integer
  uuid: string
}

export class Attachment {
  #k: string
  #crypto: Crypto
  #buffer: Buffer

  #icon?: Buffer // png buffer
  #file?: Buffer
  #metadata?: AttachmentMetadata

  private metadataSize: number
  private iconSize: number

  constructor(crypto: Crypto, k: string, buffer: Buffer) {
    this.#buffer = buffer
    this.#validate()
    this.#crypto = crypto
    this.#k = k
    this.metadataSize = buffer.readIntLE(8, 2)
    this.iconSize = buffer.readIntLE(12, 3)

    crypto.onLock(() => {
      this.#lock()
    })
  }

  /**
   * Validate attachment file.
   */
  #validate() {
    const file = this.#buffer
    invariant(
      file.slice(0, 6).toString("utf-8") === "OPCLDA",
      "Attachment must start with OPCLDA"
    )
    // @TODO: Re-enable this
    false &&
      invariant(
        file.readIntLE(7, 1) === 1,
        "The version for this attachment file format is not supported."
      )
  }

  get icon() {
    return this.#icon!
  }

  get file() {
    return this.#file!
  }

  get metadata() {
    return this.#metadata!
  }

  async unlock() {
    const crypto = this.#crypto
    const cipher = await crypto.deriveConcreteKey({ k: this.#k })
    const { metadataSize, iconSize } = this
    const buffer = this.#buffer
    this.#icon = await crypto.decryptOPData(
      buffer.slice(16 + metadataSize, 16 + metadataSize + iconSize),
      cipher
    )
    this.#file = await crypto.decryptOPData(
      buffer.slice(16 + metadataSize + iconSize),
      cipher
    )

    const metadata = JSON.parse(buffer.slice(16, 16 + metadataSize).toString("utf-8"))
    metadata.overview = JSON.parse(
      (
        await crypto.decryptOPData(
          Buffer.from(metadata.overview, "base64"),
          crypto.overview
        )
      ).toString()
    )
    this.#metadata = metadata
  }

  #lock() {
    this.#metadata = undefined!
    this.#icon = undefined!
    this.#file = undefined!
  }

  dispose() {
    this.#buffer = null!
    this.#lock()
  }
}
