import type { ItemDetails, Overview } from "../types"
import type { Crypto } from "./crypto"
import { Attachment } from "./attachment"
import { NotUnlockedError } from "../errors"
import type { Category } from "../models"

export interface EncryptedItem {
  category: string // "001"
  /** Unix seconds */
  created: integer
  d: string // "b3BkYXRhMbt"
  folder: string // 32 chars
  hmac: string // base64
  k: string // base64
  o: string // base64
  tx: integer // Unix seconds
  updated: integer // Unix seconds
  uuid: string // 32 chars
  trashed?: boolean
}

export class Item {
  #crypto: Crypto
  #data: EncryptedItem
  #overview?: Overview
  #details?: ItemDetails

  attachments: Attachment[] = []

  // Unix milliseconds
  get createdAt(): number {
    return this.#data.created * 1000
  }
  get updatedAt(): number {
    return this.#data.updated * 1000
  }
  get lastUsed() {
    return this.#data.tx
  }
  get isDeleted() {
    return this.#data.trashed
  }
  get category() {
    return this.#data.category as Category
  }
  get uuid(): string {
    return this.#data.uuid
  }
  get overview(): Overview {
    if (!this.#overview) {
      throw new NotUnlockedError()
    }
    return this.#overview!
  }
  get details(): ItemDetails {
    if (!this.#details) {
      throw new NotUnlockedError()
    }
    return this.#details!
  }

  constructor(crypto: Crypto, data: EncryptedItem) {
    this.#crypto = crypto
    this.#data = data
    crypto.onLock(() => {
      this.#overview = undefined!
      this.#details = undefined!
      this.attachments.forEach(file => file.dispose())
    })
  }

  /** @internal */
  async _unlockOverview() {
    this.#overview = await this.#crypto.decryptItemOverview(this.#data)
    return this
  }

  /** @internal */
  async _unlockDetails() {
    this.#details = await this.#crypto.decryptItemDetails(this.#data)
    return this
  }

  /** @internal */
  async _unlock() {
    await this._unlockOverview()
    await this._unlockDetails()
    return this
  }

  /** @internal */
  addAttachment(buffer: Buffer) {
    this.attachments.push(new Attachment(this.#crypto, this.#data.k, buffer))
  }
}
