import type { IAdapter } from "../adapters"
import { HMACAssertionError, invariant } from "../errors"
import { OnePasswordFileManager } from "../fs"
import { i18n } from "../i18n"
import type { EncryptedItem } from "./item"
import { Crypto } from "./crypto"
import { Item } from "./item"
import type { Profile } from "../types"
import { WeakValueMap } from "../weakMap"
import { createEventEmitter } from "../ee"

// type FoldersMap = { [uuid: string]: { [uuid: string]: Item } }

/**
 * Main OnePassword Vault class
 */
export class Vault {
  #profile: Profile
  // #folders: FoldersMap
  #items: Item[] = []
  #itemsMap = new WeakValueMap<string, Item>()
  #crypto: Crypto

  readonly onLock = createEventEmitter<void>()

  private constructor(
    profile: Profile,
    // folders: FoldersMap,
    items: Item[],
    crypto: Crypto,
    itemsMap: WeakValueMap<string, Item>
  ) {
    this.#profile = profile
    // this.#folders = folders
    this.#items = items
    this.#crypto = crypto
    this.#itemsMap = itemsMap
  }

  /**
   * Create a new OnePassword Vault instance and read all bands.
   * @internal
   */
  static async of(path: string, profileName = "default", adapter: IAdapter) {
    const crypto = new Crypto(i18n, adapter)
    const files = await OnePasswordFileManager(adapter.fs, path, profileName)
    const profile = JSON.parse(
      stripText(await files.getProfile(), /^var profile\s*=/, ";")
    )
    // const folders = JSON.parse(stripText(await files.getFolders(), "loadFolders(", ");"))

    const itemsMap = new WeakValueMap<string, Item>()
    const bands: Item[] = []

    for (let i = 0; i < 36; i++) {
      const letter = i.toString(36).toUpperCase()
      const source = await files.getBand(letter)
      if (!source) continue
      const object: Record<string, EncryptedItem> = JSON.parse(
        stripText(source, "ld(", ");")
      )
      for (const value of Object.values(object)) {
        const item = new Item(crypto, value)
        bands.push(item)
        itemsMap.set(value.uuid, item)
      }
    }

    const attachments = await files.getAttachments()
    for (const att of attachments) {
      const file = itemsMap.get(att.itemUUID)
      invariant(file, `Item ${att.itemUUID} of attachment does not exist`)
      file.addAttachment(await att.getFile())
    }

    return new Vault(profile, bands, crypto, itemsMap)
  }

  getOverview(uuid: string) {
    this.#crypto.assertUnlocked()
    return this.#items.find(x => x.uuid === uuid)?.overview
  }

  keys() {
    return this.#items.map(x => x.uuid)
  }

  async *values() {
    this.#crypto.assertUnlocked()
    for (const item of this.#items) {
      yield await item._unlock()
    }
  }

  /**
   * Unlock this OnePassword vault.
   * @param masterPassword User provided master password. Only the derived
   * master and overview key will be stored within the class.
   */
  async unlock(masterPassword: string) {
    try {
      await this.#crypto.unlock(this.#profile, masterPassword)
    } catch (e) {
      if (e instanceof HMACAssertionError) {
        throw new Error(i18n.error.invalidPassword)
      }
      throw e
    }
    return this
  }

  /**
   * Remove derived keys stored within the class instance.
   */
  lock() {
    this.#crypto.lock()
    this.onLock()
    return this
  }

  get isLocked() {
    return this.#crypto.locked
  }

  getItem(uuid: string): Promise<Item | undefined>
  getItem(filter: { title: string }): Promise<Item | undefined>

  async getItem(filter: any) {
    this.#crypto.assertUnlocked()
    if (typeof filter === "string") {
      return this.#itemsMap.get(filter)?._unlock()
    } else {
      for (const _value of this.#items) {
        const value = await _value._unlockOverview()
        if (value.overview.title === filter.title) {
          return value._unlockDetails()
        }
      }
    }
  }
}

function stripText(text: string, prefix: string | RegExp, suffix: string | RegExp) {
  if (typeof prefix === "string") {
    if (text.startsWith(prefix)) {
      text = text.slice(prefix.length)
    }
  } else {
    const prefixMatch = text.match(prefix)
    if (prefixMatch) {
      text = text.slice(prefixMatch[0].length)
    }
  }
  if (typeof suffix === "string") {
    if (text.endsWith(suffix)) {
      text = text.slice(0, -suffix.length)
    }
  } else {
    const suffixMatch = text.match(suffix)
    if (suffixMatch) {
      text = text.slice(0, -suffixMatch[0].length)
    }
  }
  return text
}
