import type { FieldType } from "./models"

type integer = number

export interface Profile {
  lastUpdatedBy: "Dropbox"
  /** Unix seconds */
  updatedAt: integer
  profileName: string
  /** base64 */
  salt: string
  /** base64 */
  masterKey: string
  iterations: integer // 50000
  uuid: string // 32 chars
  overviewKey: string // "b3B...IMO52D"
  /** Unix seconds */
  createdAt: integer
}

export type TextField = {
  type: FieldType.Text
  value: string
  designation: string
  name: string
}
export type BooleanField = {
  type: FieldType.Checkbox
  name: string
  value?: "✓" | string
}

export type ItemField =
  | TextField
  | BooleanField
  | {
      // @TODO: This currently catches all item fields.
      type: FieldType
      value: string
      designation?: string
      name: string
    }

export namespace ItemSection {
  export type A = {
    guarded: "yes"
    clipboardFilter?: string
  }

  export type Address = {
    k: "address"
    v: {
      city: string
      zip: string
      state: string
      country: string
      street: string
    }
    n: "address"
    a: {
      guarded: "yes"
    }
    t: "address"
  }

  export type String = {
    k: "string"
    v: string
    /** Unique name */
    n: string // "firstname" | "initial" | "address" | "class" | "conditions" | "expiry_date"
    a?: A
    /** User-readable title */
    t: string // "first name" | "initial" | "address" | "license class" | "conditions / restrictions" | "expiry date"
  }
  export type Menu = {
    k: "menu"
    v: string // "female"
    a: A
    t: string // "sex"
  }
  export type Date = {
    k: "date"
    v: number // 359100000
    /** @example "birthdate" */
    n: string
    a: A
    /** @example "birth date" | "date of birth" */
    t: string
  }
  export type Gender = {
    k: "gender"
    n: "sex"
    v: string // "female"
    t: "sex"
  }
  export type MonthYear = {
    k: "monthYear"
    n: string // "expiry_date"
    v: number // 2515
    t: string // "expiry date"
  }
  export type Concealed = {
    k: "concealed"
    n: "password"
    v: string
    a?: {
      generate: "off"
    }
    t: "password"
  }

  // eslint-disable-next-line @typescript-eslint/ban-types
  export type Any = String | Menu | Date | Gender | MonthYear | Concealed | Address
}

// One of them is empty?, 0C4F27910A64488BB339AED63565D148
export interface ItemDetails {
  htmlForm?: {
    /** @example "/login/" */
    htmlAction: string
    htmlMethod: "post" | "get"
  }
  notesPlain?: string
  sections?: {
    /** @example "name" | "title" | "internet" */
    name: string
    /** @example "Identification" | "Address" | "Internet Details" */
    title: string
    fields?: ItemSection.Any[]
  }[]
  /** Web form fields */
  fields?: ItemField[]
}

export interface Folder {
  created: number // 1373754128
  /** base64 */
  overview: string
  smart?: true
  tx: number // 1373754523
  /** Updated time in UNIX seconds */
  updated: number
  uuid: string // "AC78552EB06A4F65BEBF58B4D9E32080"
}

export interface Overview {
  /**
   * Password strength.
   *
   * `ps` is the value for manually entered passwords and is used to display
   * the relative strength value.
   */
  ps: number

  /**
   * Password Bits of Entropy used for generating the password.
   */
  pbe?: number

  /**
   * https://1password.community/discussion/comment/372147/#Comment_372147
   * pbe should be used for password strength display
   */
  pgrng?: boolean
  title?: string
  ainfo?: string
  url?: string
  tags?: string[]
  URLs?: { u: string }[]
  appIds: {
    /** @example 'Firefox' | 'Email' */
    name: string
    /** @example 'accessibility://login' */
    type: string
    /** @example 'android://...' */
    id: string
  }[]
}
